from staffeli import names


def _lookup_id(id, entities):
    for entity in entities:
        if entity['id'] == id:
            return entity

    ids = ["{} ({})".format(entity['id'], entity['name'])
           for entity in entities]
    raise LookupError(
        "No candidate for {}. Your options include {}.".format(
            id, ", ".join(ids)))


def _lookup_name(name, entities):
    matches = []

    for entity in entities:
        if name.lower() in entity['name'].lower():
            matches.append(entity)

    if len(matches) > 1:
        matching_names = [match['name'] for match in matches]
        raise LookupError(
            "Multiple candidates for \"{}\": {}.".format(
                name, names.pp(matching_names)))

    if len(matches) == 0:
        all_names = [entity['name'] for entity in entities]
        raise LookupError(
            "No candidate for \"{}\". Your options include {}.".format(
                name, names.pp(all_names)))

    return matches[0]


class ListedEntity:
    def __init__(self, entities=None, name=None, id=None):
        if entities is not None:
            if name is not None:
                self.json = _lookup_name(name, entities)
            elif id is not None:
                self.json = _lookup_id(id, entities)
            else:
                raise LookupError(
                    "For me to find a course, you must provide a name or id.")
        if self.json is not None:
            self.id = self.json['id']
            self.displayname = self.json['name']
        else:
            raise TypeError(
                "ListedEntity initialized with insufficient data")



#!/usr/bin/env python3

import os.path, yaml, sys, re

from canvas import Canvas

def config_dir(parent, filename):
    for i in range(5):
        if os.path.isfile(os.path.join(parent, filename)):
            return parent
        parent = os.path.join("..", parent)
    raise LookupError("Couldn't locate a {} file.".format(filename))

def dir_name(name):
    name = re.sub("[^a-zA-Z0-9-]", '_', name)
    if os.path.exists(name):
        raise Exception(
            "You already got a directory or file named '{}'.\n \
            Please rename or remove this in order to set up TA environment.".format(name))
    return name

def get_set_id(name, sets):
    for s in sets:
        if name == 'Dybber/Athas gruppe-test':
            name = 'ass1'
        if not ('name' in s and 'id' in s):
            raise LookupError("Set is missing the fields: 'name' or 'id'.")
        if s['name'] == name:
            return s['id']

    raise LookupError("No Group Set named '{}'.".format(name))

def get_members(name, ass, course_id, api_token):
    # we should probably store the group categorie id in assignment's canvas.yaml
    group_sets  = ass.canvas.group_categories(course_id)
    group_set_id = get_set_id(ass.json['name'], group_sets)

    # get ids of all groups in a group set 
    groups = ass.canvas.groups(group_set_id)
    group_id = get_set_id(name, groups)

    return Canvas(api_token).group_members(group_id)

def make_sub_dirs(name, subs):
    for i in range(len(subs)):
        sub_dir = os.path.join(name , str(i))
        os.makedirs(sub_dir)

        with open(os.path.join(sub_dir, "canvas.yaml"), "w") as f:
            yaml.dump(subs[i], f, default_flow_style=False)

def init_assignment(ass, name):

    assignment_dir = dir_name(ass)
    os.mkdir(assignment_dir)

    course_dir = config_dir(".", "canvas.yaml")    
    with open(os.path.join(course_dir, "canvas.yaml"), "r") as f:
        conf = yaml.load(f)
    course_id = conf['id']

#    with open(os.path.join(course_dir, "token"), "r") as f:
#        api_token = f.read().strip()

    assignment = Canvas().course(
        id = course_id).assignment(
            name = ass)

    subs = assignment.submissions()

    # get the ids of the students in
    # group set <assignment name>, group <instructor name>
    #members = get_members(name, assignment, course_id, api_token)
    #student_ids  = [m['id'] for m in members]

    #subs = [sub for sub in subs if sub['user_id'] in student_ids]
    make_sub_dirs(assignment_dir, subs)

    return subs


def main():
    try:
        ass = sys.argv[1]
        name = "" #sys.argv[2]
    except IndexError:
        print('error: wrong arguments', file=sys.stderr)
        print('usage: canvasTA-subs.py ASSIGNMENT_NAME',
              file=sys.stderr)
        return 1

    init_assignment(ass, name)
    print("TA environment for assignment {} was succesfully set up.".format(ass))
    return 0

if __name__ == "__main__":
    sys.exit(main())



import json
import os
import os.path
import requests


def via_post(
        api_base: str, url_relative: str, token: str, filepath: str) -> int:

    url = api_base + url_relative
    headers = {
        'Authorization': 'Bearer ' + token
    }

    name = os.path.basename(filepath)
    size = os.stat(filepath).st_size

    params = {
        'name': name,
        'size': size
    }

    resp = requests.post(url, headers=headers, params=params).json()

    upload_url = resp['upload_url']
    data = resp['upload_params']
    name = resp['file_param']

    with open(filepath, "rb") as f:
        resp = requests.post(
            upload_url, data=data, files=[(name, f)])

    return json.loads(resp.text)['id']



from staffeli import cachable
from staffeli.canvas import Course
from staffeli.listed import ListedEntity
from staffeli.submission import Submission

from typing import Any, Dict, Optional


class Assignment(ListedEntity, cachable.CachableEntity):
    def __init__(
            self, course: Course,
            name: Optional[str] = None,
            id: Optional[int] = None,
            path: Optional[str] = None) -> None:
        self.canvas = course.canvas
        self.course = course
        self.cachename = 'assignment'

        if name is None and id is None:
            if path is None:
                path = '.'
                walk = True
            else:
                walk = False
            cachable.CachableEntity.__init__(self, self.cachename, path, walk)
            ListedEntity.__init__(self)
        else:
            entities = self.canvas.list_assignments(self.course.id)
            ListedEntity.__init__(self, entities, name, id)

        self.subs = map(Submission, self.submissions())

    def publicjson(self) -> Dict[str, Any]:
        return {self.cachename: self.json}

    def submissions(self):
        return self.canvas.get(
            'courses/{}/assignments/{}/submissions?per_page=9000'.format(
                self.course.id, self.id))

    def submissions_download_url(self):
        return self.canvas.submissions_download_url(self.course.id, self.id)

    def give_feedback(self, submission_id, grade, filepaths, message,
                      use_post=False):
        self.canvas.give_feedback(
          self.course.id, self.course.displayname,
          self.id, submission_id, grade, filepaths, message, use_post)



import os
import os.path
import yaml
from typing import Any, Dict, List, Union, Tuple
from staffeli import names

STAFFELI_FILENAME = ".staffeli.yml"
TOKEN_FILENAMES = ["token", "token.txt", ".token"]
STAFFELI_RC = ".staffelirc"
MAX_DIR_SEARCH_DEPTH = 9


def _raise_lookup_error(key, attr, entities):
    all_names = [entity[attr] for entity in entities]
    raise LookupError(
        "No candidate for \"{}\". Your options include {}.".format(
            key, names.pp(all_names)))


def _raise_lookup_file(namestr, lastparent):
    raise LookupError((
            "Couldn't locate a file named {}. " +
            "I have looked for it here and in\n" +
            "parent directories up to, and including {}."
        ).format(
            namestr,
            os.path.abspath(os.path.split(lastparent)[0])))


def _find_file(
        candidate_names, parent='.', max_depth=MAX_DIR_SEARCH_DEPTH):
    if isinstance(candidate_names, str):
        candidate_names = [candidate_names]

    for i in range(MAX_DIR_SEARCH_DEPTH):
        for name in candidate_names:
            path = os.path.join(parent, name)
            if os.path.isfile(path):
                return path
        parent = os.path.join(parent, '..')

    if len(candidate_names) == 1:
        namestr = candidate_names[0]
    elif len(candidate_names) == 2:
        namestr = "either {} or {}".format(
            candidate_names[0], candidate_names[1])
    else:
        namestr = "either {}, or {}".format(
            ", ".join(candidate_names[:-1]), candidate_names[-1])

    _raise_lookup_file(namestr, parent)


def find_token_file():
    return _find_file(TOKEN_FILENAMES)


def find_rc() -> Tuple[int, str]:
    try:
        rcfile = _find_file(STAFFELI_RC)
        import configparser
        config = configparser.ConfigParser()
        config.readfp(open(rcfile))
        config = config['default']
        return (int(config['account_id']), str(config['token']))
    except LookupError:
        with open(_find_file(TOKEN_FILENAMES)) as f:
            token = f.read().strip()
        return (6, token)


def load_staffeli_file(path: str) -> Union[List[Any], Dict[Any, Any]]:
    with open(path, 'r') as f:
        return yaml.load(f)


def find_staffeli_file(
        cachename, searchdir='.', max_depth=MAX_DIR_SEARCH_DEPTH):
    parent = searchdir
    namestr = STAFFELI_FILENAME

    for i in range(max_depth):
        path = _find_file(namestr, parent, max_depth)
        model = load_staffeli_file(path)
        parent = os.path.split(path)[0]
        if len(model) == 1 and cachename in model:
            return parent, model
        parent = os.path.join(parent, '..')

    _raise_lookup_file(namestr, parent)



#!/usr/bin/env python3

import os

from assignment import get_cwd_assignment
from submission import find_student_id

def main():
    assignment = get_cwd_assignment()
    student_id = find_student_id(os.getcwd())
    print("https://absalon.ku.dk/courses/{}/gradebook/speed_grader?assignment_id={}#%7B%22student_id%22%3A%22{}%22%7D".format(
        assignment.course.id, assignment.id, student_id))

if __name__ == "__main__":
    main()



import argparse, os, os.path, shutil, yaml, sys, re

from staffeli import canvas

from urllib.request import urlretrieve

from slugify import slugify

import Levenshtein as levenshtein

from bs4 import BeautifulSoup

if os.name == "nt":
    import _winapi

def mkdir(path):
    if not os.path.exists(path):
        os.mkdir(path)

def mknewdir(path):
    if os.path.exists(path):
        raise Exception((
                "There already exists a file or directory \"{}\".\n" +
                "Please rename or remove this first."
            ).format(path))
    os.mkdir(path)

def cache(o, dirpath):
    if not os.path.isdir(dirpath):
        os.mkdir(dirpath)
    o.cache(dirpath)

def clone(args):
    dirname = " ".join(args)

    course = canvas.Canvas().course(name = dirname)
    mknewdir(dirname)
    course.cache(dirname)

    os.chdir(dirname)
    fetch_students(course)
    fetch_groups(course)

def fetch_students(course):
    print("Fetching students..")
    cache(course.list_students(), "students")

def fetch_groups(course):
    print("Fetching group categories..")
    gcs = course.list_group_categories()
    cache(gcs, "groups")

    print("Fetching group lists for each category..")
    for gc in gcs.json:
        print("Fetching {}..".format(gc['name']))
        gs = canvas.GroupList(course, id = gc['id'])
        gs.cache(os.path.join("groups", gc['name'] + ".yml"))

def fetch_group(course, name):
    gs = canvas.GroupList(course, name = name)
    mkdir("groups")
    path = os.path.join("groups", gs.name + ".yml")
    gs.cache(path)
    print("Fetched {} as {}.".format(name, path))

def fetch_all_subs(course):
    print("Fetching all assignments..")
    assigns = course.list_assignments()
    mkdir("subs")
    for assign in assigns:
        if assign['grading_type'] == 'not_graded':
            continue
        print("Fetching {}..".format(assign['name']))
        assign = canvas.Assignment(course, id = assign['id'])
        dirname = slugify(assign.displayname.replace(os.sep, '_'))
        path = os.path.join("subs", dirname)
        mkdir(path)
        assign.cache(path)

def fetch_attachments(path, attachments):
    for att in attachments:
        targetpath = os.path.join(path, att['filename'])
        print("Downloading {}..".format(targetpath))
        if os.path.isfile(targetpath):
            print("Skipped. Looks like it is already here.")
            # TODO: Do this smarter
            continue
        urlretrieve(att['url'], targetpath)

def write_body(path: str, body: str) -> None:
    with open(os.path.join(path, 'body.txt'), 'w') as f:
        soup = BeautifulSoup(body, 'html.parser')
        f.write(soup.text.strip() + "\n")

def fetch_sub(students, path, sub, metadata = False):
    json = sub.json
    sid = json['user_id']
    if (not sid in students) or \
            (not 'kuid' in students[sid]):
        print("There is something wrong with {}.. Skipping".format(sid))
        print("Looks like this is a Test Student..")
        print("Try and have a look in SpeedGrader(tm):\n{}".format(json['preview_url']))
        print(sub)
        return
    subpath = os.path.join(path, "{}_{}".format(students[sid]['kuid'], sid))
    mkdir(subpath)
    sub.cache(subpath)
    if metadata:
        return
    sub_type = json['submission_type']
    if sub_type == 'online_text_entry':
        write_body(subpath, json['body'])
    elif sub_type == 'online_upload':
        if not 'attachments' in json:
            print("There is something wrong with {}.. Skipping".format(sid))
            print("This might be a 'No submission' submission.")
            print("Try and have a look in SpeedGrader(tm):\n{}".format(json['preview_url']))
            print(sub)
            return
        fetch_attachments(subpath, json['attachments'])

def fetch_subs(course, name, deep = False, metadata = False):
    name = slugify(name)
    path = os.path.join("subs", name)
    if os.path.isdir(path):
      assign = canvas.Assignment(course, path = path)
    else:
      assign = canvas.Assignment(course, name = name)
    mkdir("subs")
    mkdir(path)
    assign.cache(path)
    print("Fetched {} as {}.".format(name, path))
    if not deep:
        return

    students = canvas.StudentList(searchdir = "students").mapping
    for sub in assign.subs:
        fetch_sub(students, path, sub, metadata)

def fetch(args, metadata):
    course = canvas.Course()
    what = args[0].rstrip(os.sep)
    args = args[1:]
    if os.sep in what:
        partition = os.path.split(what)
        what = partition[0]
        args.insert(0, partition[1])
    args = list(map(lambda s: s.rstrip(os.sep), args))
    if what == "students":
        fetch_students(course)
    elif what == "groups":
        fetch_groups(course)
    elif what == "group":
        fetch_group(course, " ".join(args))
    elif what == "subs":
        if len(args) == 0:
            fetch_all_subs(course)
        else:
            fetch_subs(course, " ".join(args), deep = True,
                metadata = metadata == True)
    else:
        raise Exception("Don't yet know how to fetch {}.".format(str(args)))

def copySub(uid, dst):
  uid = str(uid)
  for root, _, files in os.walk(subsdir):
    for f in files:
      if uid in f:
        dst = os.path.join(dst, os.path.basename(root))
        shutil.rmtree(dst, ignore_errors=True)
        shutil.copytree(root, dst, symlinks=True)
        return

def student_dirname(student):
    return "{}_{}".format(student['kuid'], student['id'])

def normalize_pathname(pathname):
    return pathname.replace(os.sep  , "_")

def split_according_to_groups(course, subspath, path):
    if not os.path.isdir(subspath):
        subspath = os.path.join("subs", subspath)
        if not os.path.isdir(subspath):
            raise LookupError("Can't resolve subs directory.")

    subsid = os.path.split(subspath)[1]

    splitpath = os.path.join("splits", subsid)
    mkdir("splits")
    mkdir(splitpath)

    gl = canvas.GroupList(course, path = path)
    teams = gl.uidmap()
    students = canvas.StudentList(searchdir = "students").mapping

    for name, uids in teams.items():
        name = normalize_pathname(name)
        namepath = os.path.join(splitpath, name)
        mkdir(namepath)
        for uid in uids:
            dirname = student_dirname(students[uid])
            subpath = os.path.join(subspath, dirname)
            if os.path.isdir(subpath):
                src = subpath
                tgt = os.path.join(namepath, dirname)
                shutil.copytree(src, tgt, symlinks=True)

def main_args_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=
"""
git-style, git-compatible command-line interface for canvas.

start a working area
    clone   Create a local clone for a course

update a working area
    fetch   Fetch something that might have changed

Grade a submission:
    grade GRADE [-m COMMENT] [-1] [FILEPATH]...

    Where
        GRADE           pass, fail, or an int.
        [-m COMMENT]    An optional comment to write.
        [-1]            Only upload feedback for 1 student id (instead of all).
        [FILEPATH]...   Optional files to upload alongside.

Work with groups:
    group add group GROUP_CATEGORY GROUP_NAME
        Add a new group to an *existing* group category.

    group set members GROUP_NAME USER...
        Set the members of a group.  Ignores any current members.

Work with users:
    user find USER_NAME
        Check if the user exists.  If not, make a guess.
""")
    parser.add_argument(
        "action", metavar="ACTION",
        help="the action to perform")
    parser.add_argument(
        "--metadata", action='store_true',
        help="fetch metadata only")
    return parser

def parse_action_arg(parser, args):
    args, remargs = parser.parse_known_args(args)
    if args.action == "help":
        parser.print_help()
        sys.exit(0)
    return args, remargs

def group(args):
    if len(args) == 4 and args[0] == 'add' and args[1] == 'group':
        add_group(args[2], args[3])

    if len(args) >= 4 and args[0] == 'set' and args[1] == 'members':
        set_group_members(args[2], args[3:])

def add_group(group_category_name, group_name):
    course = canvas.Course()

    can = canvas.Canvas()
    categories_all = can.group_categories(course.id)
    categories = list(filter(lambda x: x['name'] == group_category_name,
                             categories_all))
    if len(categories) < 1:
        raise Exception('no category of that name')
    else:
        group_category_id = categories[0]['id']

    can.create_group(group_category_id, group_name)

    fetch_groups(course) # a bit silly, and very slow

def set_group_members(group_name, user_names):
    course = canvas.Course()

    can = canvas.Canvas()
    groups_all = can.groups_in_course(course.id)
    groups = list(filter(lambda x: x['name'] == group_name,
                         groups_all))
    if len(groups) < 1:
        raise Exception('no group of that name')
    else:
        group_id = groups[0]['id']

    users_all = list(can.all_students(course.id))
    user_ids = []
    for user_name in user_names:
        users = list(filter(lambda x: x['name'] == user_name,
                            users_all))
        if len(users) < 1:
            raise Exception('no user of that name')
        else:
            user_id = users[0]['id']
        user_ids.append(user_id)
    can.add_group_members(group_id, user_ids)

    fetch_groups(course) # a bit silly, and very slow

def user(args):
    if len(args) == 2 and args[0] == 'find':
        find_user(args[1])

def find_user(user_name):
    course = canvas.Course()
    can = canvas.Canvas()
    users = list(can.all_students(course.id))

    # Hack to remove duplicates.
    users_found = list(filter(lambda user: user_name == user['name'], users))
    users_found = list(set(tuple(x.items()) for x in users_found))
    users_found = [{k: v for k, v in x} for x in users_found]
    if len(users_found) == 1:
        print('Found; id: {}, sis_user_id: {}, sis_login_id: {}'.format(
            users_found[0]['id'], users_found[0]['sis_user_id'],
            users_found[0]['sis_login_id']))
    elif len(users_found) > 1:
        print('{} users found:'.format(len(users_found)))
        for user in users_found:
            print(user)
        sys.exit(1)
    else:
        print('No users found.  Guesses:')
        users.sort(key=lambda user: levenshtein.ratio(user_name, user['name']),
                   reverse=True)
        for user in users[:10]:
            print('{} ({:2%} match)'.format(
                user['name'],
                levenshtein.ratio(user_name, user['name'])))
        sys.exit(1)

def groupsplit(args):
    split_according_to_groups(canvas.Course(), args[0], args[1])

def _check_grade(grade):
    goodgrades = ["pass", "fail"]
    if not grade in goodgrades:
        try:
            x = int(grade)
        except ValueError:
            print((
                "\"{}\" is a bad grade. Acceptable grades are: \"{}\", or an int."
                ).format(grade, "\", \"".join(goodgrades)))
            sys.exit(1)
    return grade

def _check_filepaths(filepaths):
    for filepath in filepaths:
        if not os.path.isfile(filepath):
            print("\"{}\" is not a file. What is this?".format(filepath))
            sys.exit(1)
    return filepaths

def grade_args_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="")
    parser.add_argument(
        "-m", metavar="COMMENT", dest='message', default="See attached files.")
    parser.add_argument(
        "-1", action='store_true', dest='one', default=False)
    parser.add_argument(
        "grade", metavar="GRADE")
    return parser

def grade(args):
    args, remargs = grade_args_parser().parse_known_args(args)
    grade = _check_grade(args.grade)
    filepaths = _check_filepaths(remargs)
    message = args.message

    course = canvas.Course()
    assignment = canvas.Assignment(course)
    submission = canvas.Submission()
    student_ids = submission.student_ids
    if args.one:
        student_ids = student_ids[:1]
    for sub in assignment.submissions():
        if sub['user_id'] in student_ids:
            current_grade = sub['grade']
            if current_grade == grade:
                print('Already graded.')
                course.canvas.show_verification_urls(
                    course.id, assignment.id, sub['user_id'])
                return
    for student_id in student_ids:
        assignment.give_feedback(student_id, grade,
            message, filepaths, use_post = True)

def main():
    parser = main_args_parser()
    args, remargs = parse_action_arg(parser, sys.argv[1:])
    action = args.action

    if action == "clone":
        clone(remargs)
    elif action == "fetch":
        fetch(remargs, args.metadata)
    elif action == "grade":
        grade(remargs)
    elif action == "group":
        group(remargs)
    elif action == "user":
        user(remargs)
    elif action == "groupsplit":
        groupsplit(remargs)
    else:
        print("Unknown action {}.".format(action))
        parser.print_usage()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os, os.path, sys, yaml

def find_student_ids(subdir):
    jsonpath = os.path.join(subdir, "canvas_group.json")
    if os.path.isfile(jsonpath):
        with open(jsonpath, "r") as f:
            return json.load(f)
    else:
        return [find_student_id(subdir)]

def find_student_id(subdir):
    yamlpath = os.path.join(subdir, "canvas.yaml")
    if os.path.isfile(yamlpath):
        with open(yamlpath, "r") as f:
            return yaml.load(f)["user_id"]

    for filename in os.listdir(subdir):
        if filename.count("_") >= 3:
            filename = filename.replace("_late_", "_")
            parts = filename.split("_")
            return int(parts[1])
    raise LookupError("Can't find student id in \"{}\".".format(subdir))

def main():
    print(find_student_ids("."))

if __name__ == "__main__":
    main()



import os.path
import yaml
from staffeli import files

from typing import Any


class CachableEntity:
    cachename = None  # type: str

    def __init__(
            self, cachename: str,
            path: str = '.', walk: bool = True
            ) -> None:
        if os.path.isdir(path):
            depth = [i for i in [1] if not walk]
            self.parentdir, model = files.find_staffeli_file(
                cachename, path, *depth)
        elif os.path.isfile(path):
            self.parentdir = os.path.split(path)[0]
            model = files.load_staffeli_file(path)
        else:
            raise LookupError((
                "{} is neither a directory, nor file path."
                ).format(path))
        self.json = model[cachename]

    def cache(self, path: str) -> None:
        if os.path.isdir(path):
            path = os.path.join(path, ".staffeli.yml")
        with open(path, 'w') as f:
            yaml.dump(
                self.publicjson(),
                f, default_flow_style=False, encoding='utf-8')

    def publicjson(self) -> Any:
        return self.json



#!/usr/bin/env python3

import os, os.path, re, shutil, sys, yaml

from assignment import get_cwd_assignment
from submission import find_student_id

import datetime

def _iso8601(datestr):
    return datetime.datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%SZ")

def get_subdirs():
    parent = "."
    subs = {}
    for name in os.listdir(parent):
        path = os.path.join(parent, name)

        if not os.path.isdir(path):
            continue

        try:
            subs[find_student_id(path)] = name
        except LookupError:

            # Hotfix for ill-formed AP subs
            for subdir in os.listdir(path):
                path = os.path.join(path, subdir)
                if os.path.isdir(path):
                    subs[find_student_id(path)] = name
    return subs

def get_resubs(assignment):
    resubs = []
    for sub in assignment.submissions():
        if sub["grade_matches_current_submission"] == False:
            resubs.append(sub)
    return resubs

def download_resub(resub, subdir):
    dst = os.path.join(subdir, "resub" + str(resub['attempt']))
    shutil.rmtree(dst, ignore_errors=True)
    os.mkdir(dst)

def mkdirp(path):
    if os.path.isdir(path):
        return
    if os.path.isfile(path):
        raise Exception("Can't create directory {}: path refers to a file.")
    os.mkdir(path)

def grader_ids(resubs):
    grader_ids = set()
    for resub in resubs:
        grader_ids.add(resub['grader_id'])
    return grader_ids

def grader_names(canvas, grader_ids):
    names = {}
    for grader_id in grader_ids:
        name = canvas.user(grader_id)['name']
        name = re.sub('[^a-z0-9]', '', name, flags=re.IGNORECASE).lower()
        name = "{}_{}".format(name, grader_id)
        names[grader_id] = name
    return names

def download_last_graded(canvas, user_id, history, last_graded_path):
    sub_history = history['submission_history']
    last_graded = None
    for sub in sub_history:
        if sub['workflow_state'] == "graded":
            if last_graded == None or \
                    _iso8601(last_graded['graded_at']) < _iso8601(sub['graded_at']):
                last_graded = sub

    if last_graded == None:
        if len(sub_history) > 1 or not 'attachments' in sub_history[0]:
            print("Can't deduce last graded submission for {}.".format(user_id))
            return
        else:
            last_graded = sub_history[0]

    for attachment in last_graded['attachments']:
        path = os.path.join(last_graded_path, attachment['filename'])
        canvas.get_verified_file(path, attachment['url'])

def download_last_comment(canvas, user_id, history, last_graded_path):
    sub_comments = history['submission_comments']
    last_comment = None
    for comment in sub_comments:
        if not 'attachments' in comment or len(comment['attachments']) == 0:
            continue
        if last_comment == None or \
                _iso8601(last_comment['created_at']) < _iso8601(comment['created_at']):
            last_comment = comment

    if last_comment == None:
        if len(sub_comments) > 1 or not 'attachments' in sub_comments[0]:
            print("Can't deduce last comment for {}.".format(user_id))
            return
        else:
            last_comment = sub_comments[0]

    for attachment in last_comment['attachments']:
        path = os.path.join(last_graded_path, attachment['filename'])
        canvas.get_verified_file(path, attachment['url'])

def download_resub(assignment, resubsbase, graders, resub):
    canvas = assignment.canvas
    user_id = resub['user_id']
    dirpath = os.path.join(resubsbase,
            graders[resub['grader_id']],
            str(user_id)   # TODO: Use a better name.
        )
    mkdirp(dirpath)
    print("Downloading {}".format(dirpath))

    for attachment in resub['attachments']:
        path = os.path.join(dirpath, attachment['filename'])
        canvas.get_verified_file(path, attachment['url'])

    canvas_path = os.path.join(dirpath, "canvas.yaml")
    with open(canvas_path, 'w') as outfile:
        yaml.dump(resub, outfile)

    history = canvas.submission_history(assignment.course.id, assignment.id, user_id)
    history_path = os.path.join(dirpath, "history.yaml")
    with open(history_path, 'w') as outfile:
        yaml.dump(history, outfile)

    last_graded_path = os.path.join(dirpath, "last_graded")
    mkdirp(last_graded_path)

    download_last_graded(canvas, user_id, history, last_graded_path)
    download_last_comment(canvas, user_id, history, last_graded_path)

def download_all(assignment, resubs, subdirs):
    basename = os.path.basename(os.getcwd())

    resubsdir = os.path.join("..", "..", "resubs")
    mkdirp(resubsdir)

    resubsbase = os.path.join(resubsdir, basename)
    mkdirp(resubsbase)

    graders = grader_names(assignment.canvas, grader_ids(resubs))
    for grader in graders.values():
        graderbase = os.path.join(resubsbase, grader)
        mkdirp(graderbase)

    for resub in resubs:
        download_resub(assignment, resubsbase, graders, resub)

def main():
    assignment = get_cwd_assignment()
    course = assignment.course
    resubs = get_resubs(assignment)
    subdirs = get_subdirs()
    download_all(assignment, resubs, subdirs)

if __name__ == "__main__":
    main()



from staffeli import listed, cachable
from staffeli.typed_canvas import Canvas

from typing import Any, Optional


class Course(listed.ListedEntity, cachable.CachableEntity):
    canvas = None  # type: Canvas
    id = None  # type: int
    displayname = None  # type: str

    def __init__(
        self,
        canvas: Optional[Canvas]=None,
        name: Optional[str]=None,
        id: Optional[int]=None
            ) -> None:

        if canvas is None:
            self.canvas = Canvas()
        else:
            self.canvas = canvas

        self.cachename = 'course'

        if name is None and id is None:
            cachable.CachableEntity.__init__(self, self.cachename)
            listed.ListedEntity.__init__(self)
        else:
            entities = self.canvas.list_courses()
            listed.ListedEntity.__init__(self, entities, name, id)

        assert isinstance(self.json['id'], int)
        assert isinstance(self.json['name'], str)

        self.id = self.json['id']
        self.displayname = self.json['name']

    def web_url(self) -> str:
        return self.canvas.web_url('courses/{}/'.format(self.id))

    def create_section(self, name: str) -> Any:
        return self.canvas.create_section(self.id, name)

    def list_sections(self) -> Any:
        return self.canvas.list_sections(self.id)

    def delete_section(self, section_id: int) -> Any:
        return self.canvas.delete_section(section_id)

    def section_enroll(self, section_id: int, user_id: int) -> Any:
        return self.canvas.section_enroll(section_id, user_id)

    def create_group_category(self, name: str) -> Any:
        return self.canvas.create_group_category(self.id, name)

    def list_group_categories(self) -> Any:
        return self.canvas.list_group_categories(self.id)

    def delete_group_category(self, gcat_id: int) -> Any:
        return self.canvas.delete_group_category(gcat_id)



#!/usr/bin/env python3

import copy
import json
import os
import requests
import sys
import time
import urllib.parse
import urllib.request
import yaml
import pdb

from os.path import basename

from staffeli import cachable, files, listed, names, upload

def format_json(d):
    return json.dumps(d, sort_keys=True, indent=2, ensure_ascii=False)

def _api_bool(value: bool) -> int:
    if value:
        return 1
    else:
        return 0

def _req(token, method, api_base, url_relative, url_absolute=None, **args):
    try:
        args = args['_arg_list']
    except KeyError:
        pass
    if type(args) == type({}):
        args = list(args.items())

    # In the case of list-returning API calls, maximize the number of entries
    # returned.  100 appears to be the max in at least one instance.  Combine
    # this with the 'all_pages=True' argument in calling '_call_api'.
    args.append(('per_page', 100))

    query_string = urllib.parse.urlencode(args, safe='[]@', doseq=True).encode('utf-8')

    if url_absolute is not None:
        url = url_absolute
    else:
        url = api_base + url_relative
    headers = {
        'Authorization': 'Bearer ' + token
    }
    return urllib.request.Request(url, data=query_string, method=method,
                                 headers=headers)

def _parse_pagination_link(s):
    link, rel = s.split('; rel="')
    link = link[1:-1]
    rel = rel[:-1]
    return (rel, link)

def _call_api(token, method, api_base, url_relative, all_pages=False, **args):
    req = _req(token, method, api_base, url_relative, None, **args)
    entries = []
    while True:
        with urllib.request.urlopen(req) as f:
            data = json.loads(f.read().decode('utf-8'))
            if type(data) is list:
                entries.extend(data)
            else:
                entries.append(data)

            # In some cases we want to extract many entries, e.g. the students
            # in a course.  However, some Absalon instances set a per_page limit
            # to 100, so we cannot just set per_page to 9000 and hope for the
            # best.  Instead we utilize the API's pagination facilities
            # documented at
            # <https://canvas.instructure.com/doc/api/file.pagination.html>.
            # This works, although it is not foolproof in the extreme case that
            # entries are added or removed from Absalon between our requests.
            # This is probably not something to worry about.
            if all_pages:
                pagination_links = {rel: link
                                    for rel, link in
                                    (_parse_pagination_link(s)
                                     for s in f.info()['Link'].split(','))}
                if pagination_links['current'] == pagination_links['last']:
                    break
                else:
                    url_absolute = pagination_links['next']
                    req = _req(token, method, api_base, None, url_absolute, **args)
            else:
                break
    if len(entries) == 1 and all_pages == False:
        return entries[0]
    else:
        return entries

def _upload_transit(course, filepath):
    form_url = "https://file-transit.appspot.com/upload"
    params = {
        'course': course
    }
    with open(filepath, "rb") as f:
        resp = requests.post(
            form_url, params=params, files=[('file', f)])
    if resp.status_code != 200:
        raise Exception(
            "Something is wrong with the file-transit service :-( " +
                resp.headers)

    print("Transitting {} to Canvas via {}".format(filepath, resp.url))

    return resp.url

def _upload_via_url(token, api_base, url_relative, filepath, viaurl):

    url = api_base + url_relative
    headers = {
        'Authorization': 'Bearer ' + token
    }

    name = basename(filepath)
    size = os.stat(filepath).st_size

    params = {
        'url'     : viaurl,
        'name'    : name,
        'size'    : size
    }

    resp = requests.post(url, headers=headers, params=params)

    json = resp.json()

    id = json['id']
    status_url = json['status_url']

    while json['upload_status'] == 'pending':
      print("Waiting for Canvas to download it..")
      time.sleep(3)
      json = requests.get(status_url, headers=headers).json()

    if json['upload_status'] != 'ready':
        raise Exception("Canvas refused to upload the file(s):\n{}". format(json))

    print("Canvas got it!")

    attachment = json['attachment']

    return attachment['id']

def _upload_submission_comment_file(
        token, api_base, url_relative, course, filepath, use_post = False):
    if use_post:
        return upload.via_post(
            api_base,
            url_relative + "/comments/files",
            token,
            filepath)
    else:
        viaurl = _upload_transit(course, filepath)
        return _upload_via_url(
            token, api_base,
            url_relative + "/comments/files",
            filepath, viaurl)

class GroupList(listed.ListedEntity, cachable.CachableEntity):
    def __init__(self, course, path = None, name = None, id = None):
        self.cachename = 'groups'
        self.canvas = course.canvas

        if path != None:
            cachable.CachableEntity.__init__(self, self.cachename, path, walk = False)
        else:
            if id != None:
                self.id = id
            else:
                entities = self.canvas.group_categories(course.id)
                listed.ListedEntity.__init__(self, entities, name)

            self.json = self.canvas.groups(self.id)

        for group in self.json:
            group['members'] = self.canvas.group_members(group['id'])

    def uidmap(self):
        mapping = {}
        for group in self.json:
            name = group['name']
            mapping[name] = map(lambda m: m['id'], group['members'])
        return mapping

    def publicjson(self):
        return { self.cachename : self.json }

class GroupCategoryList(cachable.CachableEntity):
    def __init__(self, canvas, course_id):
        self.json = canvas.group_categories(course_id)

    def publicjson(self):
        json = copy.deepcopy(self.json)
        for cat in json:
            del cat['is_member']
        return { 'group_categories': json }

class Course(listed.ListedEntity, cachable.CachableEntity):
    def __init__(self, canvas = None, name = None, id = None):

        if canvas == None:
            canvas = Canvas()

        self.canvas = canvas
        self.cachename = 'course'

        if name == None and id == None:
            cachable.CachableEntity.__init__(self, self.cachename)
            listed.ListedEntity.__init__(self)
        else:
            entities = self.canvas.courses()
            listed.ListedEntity.__init__(self, entities, name, id)

        self.id = self.json['id']
        self.displayname = self.json['name']

    def assignment(self, name = None, id = None):
        return Assignment(self, name, id)

    def list_assignments(self):
        return self.canvas.list_assignments(self.id)

    def list_students(self):
        return StudentList(self)

    def list_group_categories(self):
        return GroupCategoryList(self.canvas, self.id)

    def group(self, name = None, id = None):
        return Group(self.canvas, name, id)

    def publicjson(self):
        json = copy.deepcopy(self.json)
        del json['enrollments']
        return { self.cachename : json }

class StudentList(cachable.CachableEntity):
    def __init__(self, course = None, searchdir = "."):
        self.cachename = 'students'

        if course == None:
            cachable.CachableEntity.__init__(self, self.cachename, searchdir)
        else:
            self.json = course.canvas.all_students(course.id)

        self.mapping = {}
        for student in self.json:
            sid = student['id']
            self.mapping[sid] = student
            if 'login_id' in student:
                self.mapping[sid]['kuid'] = student['login_id'][:6]
            # 'sis_login_id' is a deprecated field according to
            # <https://canvas.instructure.com/doc/api/users.html>.
            elif 'sis_login_id' in student:
                self.mapping[sid]['kuid'] = student['sis_login_id'][:6]

    def publicjson(self):
        return { self.cachename : self.json }

class Submission(cachable.CachableEntity):
    def __init__(self, json = None):
        self.cachename = 'submission'
        if json == None:
            cachable.CachableEntity.__init__(self, self.cachename)
        else:
            self.json = json

        if 'student_ids' in self.json:
            self.student_ids = self.json['student_ids']
        else:
            self.student_ids = [self.json['user_id']]

    def publicjson(self):
        return { self.cachename : self.json }

class Assignment(listed.ListedEntity, cachable.CachableEntity):
    def __init__(self, course, name = None, id = None, path = None):
        self.canvas = course.canvas
        self.course = course
        self.cachename = 'assignment'

        if name == None and id == None:
            if path == None:
                path = '.'
                walk = True
            else:
                walk = False
            cachable.CachableEntity.__init__(self, self.cachename, path, walk)
            listed.ListedEntity.__init__(self)
        else:
            entities = self.canvas.list_assignments(self.course.id)
            listed.ListedEntity.__init__(self, entities, name, id)

        self.subs = list(map(Submission, self.submissions()))

    def publicjson(self):
        return { self.cachename : self.json }

    def submission(self, id):
        return self.canvas.get(
            'courses/{}/assignments/{}/submissions/{}'.format(
                self.course.id, self.id, id))

    def submissions(self):
        return self.canvas.get(
            'courses/{}/assignments/{}/submissions'.format(
                self.course.id, self.id),
            all_pages=True)

    def submissions_download_url(self):
        return self.canvas.submissions_download_url(self.course.id, self.id)

    def give_feedback(self, submission_id, grade, filepaths, message,
        use_post = False):
        self.canvas.give_feedback(
          self.course.id, self.course.displayname,
          self.id, submission_id, grade, filepaths, message, use_post)

def _raise_lookup_file(namestr, lastparent):
    raise LookupError((
            "Couldn't locate a file named {}. " +
            "I have looked for it here and in\n" +
            "parent directories up to, and including {}."
        ).format(
            namestr,
            os.path.abspath(os.path.split(lastparent)[0])))

class Canvas:
    def __init__(self,
                 token=None,
                 account_id=None,
                 api_base='https://absalon.ku.dk/api/v1/'):
        self.api_base = api_base

        if token is None:
            (account_id, token) = files.find_rc()

        self.account_id = account_id
        self.token = token

    def get(self, url_relative, **args):
        return _call_api(self.token, 'GET', self.api_base, url_relative, **args)

    def get_verified_file(self, path, url):
        return urllib.request.urlretrieve(url, filename=path)

    def post(self, url_relative, **args):
        return _call_api(self.token, 'POST', self.api_base, url_relative, **args)

    def put(self, url_relative, **args):
        return _call_api(self.token, 'PUT', self.api_base, url_relative, **args)

    def delete(self, url_relative, **args):
        return _call_api(self.token, 'DELETE', self.api_base, url_relative, **args)

    def course(self, name = None, id = None):
        return Course(self, name, id)

    def courses(self):
        return self.get('courses')

    def get_course(self, course_id):
        return self.get('courses/{}'.format(course_id))

    def course_create(self, name: str, license: str = 'private', is_public: bool = False):
        courses = self.courses()
        for c in courses:
            if name == c['name']:
                raise Exception(
                    "The course {} already exists. YAML dump:\n{}".format(
                        name, yaml.dump(c, default_flow_style=False)))
        return self.post('accounts/6/courses', _arg_list=[
                ('course[name]', name),
                ('course[course_code]', ''),
                ('course[license]', license),
                ('course[is_public]', _api_bool(is_public)),
                ('enroll_me', 'true')
            ])

    def list_sections(self, course_id):
        return self.get('courses/{}/sections'.format(course_id), all_pages=True)

    def create_section(self, course_id, name):
        sections = self.list_sections(course_id)
        existing = [s for s in sections if name == s['name']]
        for s in existing:
            raise Exception(
                "The section {} already exists. YAML dump:\n{}".format(
                    name, yaml.dump(s, default_flow_style=False)))
        _arg_list = [('course_section[name]', name)]
        return self.post('courses/{}/sections'.format(course_id),
            _arg_list=_arg_list)

    def delete_section(self, section_id):
        return self.delete('sections/{}'.format(section_id))

    def section_enroll(self, section_id, user_id):
        _arg_list = [('enrollment[user_id]', user_id)]
        return self.post('sections/{}/enrollments'.format(section_id),
            _arg_list=_arg_list)

    def all_students(self, course_id):
        sections = self.get('courses/{}/sections'.format(course_id),
                            _arg_list=[('include','students')])
        students = []
        for section in sections:
            if 'students' in section:
                students.extend(section['students'])

        return students

    def user(self, user_id):
        return self.get('users/{}/profile'.format(user_id))

    def course_student(self, course_id, user_id):
        user = self.get('courses/{}/users/{}'.format(course_id, user_id))
        return user

    def group_categories(self, course_id):
        return self.get('courses/{}/group_categories'.format(course_id))

    def create_group_category(self, course_id: int, name: str):
        return self.post(
            'courses/{}/group_categories'.format(course_id),
            name=name)

    def list_group_categories(self, course_id: int):
        return self.get(
            'courses/{}/group_categories'.format(course_id),
            all_pages=True)

    def delete_group_category(self, gcat_id: int):
        return self.delete(
            'group_categories/{}'.format(gcat_id))

    ########## group methods ######################
    def groups(self, group_category_id):
        return self.get('group_categories/{}/groups'.format(group_category_id),
                        all_pages=True)
    def groups_in_course(self, course_id):
        return self.get('courses/{}/groups'.format(course_id),
                        all_pages=True)
    def group(self, group_id):
        return self.get('/groups/{}'.format(group_id))

    def group_members(self, group_id):
        return self.get('/groups/{}/users'.format(group_id),
                        all_pages=True)
    ##################################

    def create_group(self, group_category_id, name):
        return self.post('group_categories/{}/groups'.format(group_category_id),
                         name=name, join_level='invitation_only')

    def delete_all_assignment_groups(self, group_category_id):
        groups = self.get('group_categories/{}/groups'.format(group_category_id),
                          all_pages=True)
        group_ids = [g['id'] for g in groups]
        for gid in group_ids:
            self.delete('groups/{}'.format(gid))

    def add_group_members(self, group_id, members):
        args = {
            'members[]': members
        }
        return self.put('groups/{}'.format(group_id), **args)

    def list_assignments(self, course_id):
        return self.get('courses/{}/assignments'.format(course_id))

    def assignment(self, course_id, assignment_id):
        return self.get(
            'courses/{}/assignments/{}'.format(
                course_id, assignment_id))

    def submission_history(self, course_id, assignment_id, student_id):
        args = [
            ('include[]', 'visibility'),
            ('include[]', 'submission_history'),
            ('include[]', 'submission_comments'),
            ('include[]', 'rubric_assessment')
        ]
        url = 'courses/{}/assignments/{}/submissions/{}'.format(
            course_id, assignment_id, student_id)
        return self.get(url, _arg_list=args)

    def submissions_download_url(self, course_id, assignment_id):
        return self.assignment(
            course_id, assignment_id)['submissions_download_url']

    def give_feedback(self,
            course_id, course_name, assignment_id, user_id, grade, message, filepaths,
            use_post = False):

        url_relative = \
            'courses/{}/assignments/{}/submissions/{}'.format(
                course_id, assignment_id, user_id)

        upload = lambda filepath : _upload_submission_comment_file(
            self.token, self.api_base, url_relative, course_name, filepath, use_post)
        ids = list(map(upload, filepaths))

        _arg_list = list(map(lambda x: ("comment[file_ids][]", x), ids))
        _arg_list.append(("comment[text_comment]", message))
        _arg_list.append(("comment[group_comment]", True))
        _arg_list.append(("submission[posted_grade]", grade))

        resp = self.put(url_relative, _arg_list=_arg_list)
        if not 'grade' in resp:
          raise Exception("Canvas response looks weird: {}".format(resp))

        print("Looks good.")
        self.show_verification_urls(course_id, assignment_id, user_id)
        return resp

    def show_verification_urls(self, course_id, assignment_id, user_id):
        speedgrader_url = "https://absalon.ku.dk/courses/{}/gradebook/speed_grader?assignment_id={}#%7B%22student_id%22%3A%22{}%22%7D".format(course_id, assignment_id, user_id)
        gradebook_url = "https://absalon.ku.dk/courses/{}/assignments/{}/submissions/{}/".format(course_id, assignment_id, user_id)

        print("Verification URLs: \n{}\n{}".format(
            speedgrader_url, gradebook_url))

def main(args):
    try:
        method = args[0].upper()
        url = args[1]
        args = args[2:]
        assert len(args) % 2 == 0
        args = [(args[i], args[i + 1]) for i in range(0, len(args), 2)]
    except IndexError:
        print('error: wrong arguments', file=sys.stderr)
        print('usage: canvas.py [GET|POST|PUT] URL [ARG_NAME ARG_VALUE]...',
              file=sys.stderr)
        return 1

    c = Canvas()
    call = {
        'GET': c.get,
        'POST': c.post,
        'PUT': c.put
    }[method]

    output = call(url, _arg_list=args)
    print(format_json(output))
    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import os
import sys

from setuptools import setup
from typing import List

if sys.version_info < (3, 3):
    sys.exit("Staffeli requires Python >= 3.3.")


def syblingpath(fname: str) -> str:
    return os.path.join(os.path.dirname(__file__), fname)


def read(fname: str) -> str:
    return open(syblingpath(fname)).read()


def readlines(fname: str) -> List[str]:
    with open(syblingpath(fname)) as f:
        return [line for line in f]


setup(
    name="staffeli",
    description="DIKU support tools for Canvas LMS",
    install_requires=[
        "pyyaml>=3.12",
        "requests>=2.12.5",
        "python-slugify>=1.2.1",
        "sphinx>=1.5.2",
        "python-Levenshtein>=0.11.0",
        "beautifulsoup4>=4.5.3",
        "typing>=3.6.1"
    ],
    tests_require=readlines("test-requirements.txt"),
    keywords=["staffeli", "Canvas", "DIKU"],
    long_description=read('README.rst'),
    url="https://github.com/DIKU-EDU/staffeli",
    version='0.4.2',
    packages=["staffeli"],
    maintainer="Oleks",
    maintainer_email="oleks@oleks.info",
    license="EUPLv1.1",
    entry_points={
        'console_scripts': [
            'staffeli = staffeli.cli:main'
        ]
    },
)



import json
import urllib
import urllib.parse
import urllib.request

from typing import Dict  # noqa: F401
from typing import Any, BinaryIO, List, Optional, Tuple, Union
from urllib.request import Request
from http.client import HTTPResponse

from staffeli import files

QueryArg = Union[int, str]


def _req(token: str, method: str, url: str, **args: QueryArg) -> Request:
    """The root of staffeli: Perform an API request.

    Handle with care."""
    query_string = urllib.parse.urlencode(
        list(args.items()),
        safe='[]@', doseq=True).encode('utf-8')

    headers = {'Authorization': 'Bearer ' + token}

    return urllib.request.Request(
        url, data=query_string, method=method, headers=headers)


def _read_json(f: Union[HTTPResponse, BinaryIO]) -> Any:
    return json.loads(f.read().decode('utf-8'))


def _list_req(token: str, method: str, url: str, **args: QueryArg) -> Request:
    # In the case of list-returning API calls, maximize the number of entries
    # returned.  100 appears to be the max in at least one instance.
    args['per_page'] = 100
    return _req(token, method, url, **args)


def _parse_pagination_link(s: str) -> Tuple[str, str]:
    link, rel = s.split('; rel="')
    link = link[1:-1]
    rel = rel[:-1]
    return (rel, link)


def _api(
        token: str, method: str, url: str,
        **args: QueryArg) -> Any:
    req = _req(token, method, url, **args)
    with urllib.request.urlopen(req) as f:
        assert isinstance(f, HTTPResponse)
        return _read_json(f)


def _list_api(
        token: str, method: str, url: str,
        **args: QueryArg) -> List[Any]:
    req = _list_req(token, method, url, **args)
    entries = []  # type: List[Any]
    while True:
        with urllib.request.urlopen(req) as f:
            assert isinstance(f, HTTPResponse)

            data = _read_json(f)
            if type(data) is list:
                entries.extend(data)
            else:
                entries.append(data)

            # In some cases we want to extract many entries, e.g. the students
            # in a course.  However, some Absalon instances set a per_page
            # limit to 100, so we cannot just set per_page to 9000 and hope
            # for the best.  Instead we utilize the API's pagination facilities
            # documented at
            # <https://canvas.instructure.com/doc/api/file.pagination.html>.
            # This works, although it is not foolproof in the extreme case that
            # entries are added or removed from Absalon between our requests.
            # This is probably not something to worry about.
            messages = str(f.getheader('Link')).split(',')
            links = [_parse_pagination_link(m) for m in messages]
            pagination_links = {rel: link for rel, link in links}
            if pagination_links['current'] == pagination_links['last']:
                break
            else:
                url = pagination_links['next']
                req = _list_req(token, method, url, **args)
    return entries


def _api_bool(value: bool) -> int:
    if value:
        return 1
    else:
        return 0


class Canvas:
    base_url = None  # type: str
    api_base = None  # type: str

    def __init__(
        self,
        token: Optional[str]=None,
        account_id: Optional[int]=None,
        base_url: str='https://absalon.ku.dk/',
        api_base: str='api/v1/'
            ) -> None:

        self.base_url = base_url
        self.api_base = api_base

        if token is None or account_id is None:
            self.account_id, self.token = files.find_rc()
        else:
            self.account_id = account_id  # type: int
            self.token = token  # type: str

    def api_url(self, rel_url: str) -> str:
        return self.base_url + self.api_base + rel_url

    def web_url(self, rel_url: str) -> str:
        return self.base_url + rel_url

    def get_list(self, rel_url: str, **args: QueryArg) -> List[Any]:
        return _list_api(self.token, 'GET', self.api_url(rel_url), **args)

    def post(self, rel_url: str, **args: QueryArg) -> List[Any]:
        return _api(self.token, 'POST', self.api_url(rel_url), **args)

    def delete(self, rel_url: str, **args: QueryArg) -> List[Any]:
        return _api(self.token, 'DELETE', self.api_url(rel_url), **args)

    def list_courses(self) -> List[Any]:
        url = 'courses'
        return self.get_list(url)

    def create_course(
            self,
            name: str,
            license: str = 'private',
            is_public: bool = False) -> Any:
        url = 'accounts/{}/courses'.format(self.account_id)
        args = {
            'course[name]': name,
            'course[course_code]': '',
            'course[license]': license,
            'course[is_public]': _api_bool(is_public),
            'enroll_me': 'true'
        }  # type: Dict[str, Union[int, str]]
        return self.post(url, **args)

    def list_sections(self, course_id: int) -> List[Any]:
        url = 'courses/{}/sections'.format(course_id)
        return self.get_list(url)

    def create_section(self, course_id: int, name: str) -> Any:
        url = 'courses/{}/sections'.format(course_id)
        args = {'course_section[name]': name}
        return self.post(url, **args)

    def delete_section(self, section_id: int) -> Any:
        """Delete a section.

        Note how this method does not take a course ID.
        Section IDs are global across all courses, and,
        to our knowledge, are not instantly re-usable."""
        url = 'sections/{}'.format(section_id)
        return self.delete(url)

    def section_enroll(self, section_id: int, user_id: int) -> Any:
        url = 'sections/{}/enrollments'.format(section_id)
        args = {'enrollment[user_id]': user_id}
        return self.post(url, **args)

    def create_group_category(self, course_id: int, name: str) -> Any:
        url = 'courses/{}/group_categories'.format(course_id)
        args = {'name': name}
        return self.post(url, **args)

    def list_group_categories(self, course_id: int) -> List[Any]:
        url = 'courses/{}/group_categories'.format(course_id)
        return self.get_list(url)

    def delete_group_category(self, gcat_id: int) -> Any:
        """Delete a group category.

        Note how this method does not take a course ID.
        Group Category IDs are global across all courses, and,
        to our knowledge, are not instantly re-usable."""
        url = 'group_categories/{}'.format(gcat_id)
        return self.delete(url)
